"""Define functions for the interface with the user."""
from __future__ import annotations

from typing import Optional, Type

import inquirer
from jinja2 import Template

from scicookie.logs import SciCookieErrorType, SciCookieLogs


def _create_question(
    question_id: str, question: dict
) -> Optional[inquirer.questions.Question]:
    # validation
    if not question.get("enabled", False):
        return None

    # config required
    question_type = question.get("type", "")
    # todo: implement help text int he prompt
    # question_help = question.get("help")
    # todo: implement depends on workflow, it needs refactoring the code
    #       because it needs access to the answer
    # question_depends_on = question.get("depends_on")

    question_types_available = [
        "text",
        "single-choice",
        "multiple-choices",
        "confirm",
    ]

    if question_type not in question_types_available:
        SciCookieLogs.raise_error(
            f"The question type ({question_type}) is not available ("
            f"{question_types_available}).",
            SciCookieErrorType.SCICOOKIE_INVALID_CONFIGURATION,
        )

    content = {"message": ""}

    if question.get("choices"):
        content["choices"] = question.get("choices", [])

    fn_questions: dict[str, Type[inquirer.questions.Question]] = {
        "text": inquirer.Text,
        "single-choice": inquirer.List,
        "multiple-choices": inquirer.Checkbox,
        "confirm": inquirer.Confirm,
    }

    return fn_questions[question_type](
        question_id,
        **content,
    )


def make_questions(questions: dict):
    """Generate all the enabled questions."""
    answers: dict[str, str] = {}

    for question_id, question in questions.items():
        question_obj = _create_question(question_id, question)
        # note: if question_object is None, that means that the question is
        #       not enabled
        if question_obj:
            default_answer = question.get("default", "")
            default_answer = Template(default_answer).render(answers)
            message = question.get("message", "")
            print(f"{message} (default: {default_answer}):")
            print(">> HELP:", question["help"])
            answer = inquirer.prompt([question_obj])

            # note: if answer is none, it means that the user cancelled
            #       the process.
            if answer is None:
                return {}
            answers[question_id] = (
                answer.get(question_id, "") or default_answer
            )
    return answers



"""Module for functions and classes for systen logs."""
import os
from enum import Enum

from colorama import Fore


class SciCookieErrorType(Enum):
    """Enum class for the error codes."""

    SH_ERROR_RETURN_CODE = 1
    SH_KEYBOARD_INTERRUPT = 2
    SCICOOKIE_INVALID_PARAMETER = 3
    SCICOOKIE_MISSING_PARAMETER = 4
    SCICOOKIE_INVALID_CONFIGURATION = 5


class SciCookieLogs:
    """Main log for print and raise messages to the system."""

    @staticmethod
    def raise_error(message: str, message_type: SciCookieErrorType):
        """Print an error message and exit with the given error code."""
        print(Fore.RED, f"[EE] {message}", Fore.RESET)
        os._exit(message_type.value)

    @staticmethod
    def info(message: str):
        """Print an info message."""
        print(Fore.BLUE, message, Fore.RESET)

    @staticmethod
    def warning(message: str):
        """Print a warning message."""
        print(Fore.YELLOW, message, Fore.RESET)



"""Module with CLI functions."""
import json
import os
import sys
from pathlib import Path
from typing import Union

import sh

from scicookie.logs import SciCookieErrorType, SciCookieLogs
from scicookie.profile import Profile
from scicookie.ui import make_questions

PACKAGE_PATH = Path(__file__).parent
COOKIECUTTER_FILE_PATH = PACKAGE_PATH / "cookiecutter.json"


def _get_cookiecutter_default_answer(
    answer_definition: Union[str, list]
) -> str:
    if not isinstance(answer_definition, (str, list)):
        SciCookieLogs.raise_error(
            "Invalid cookiecutter configuration.",
            SciCookieErrorType.SCICOOKIE_INVALID_CONFIGURATION,
        )

    if isinstance(answer_definition, str):
        return answer_definition

    return answer_definition[0]


def call_cookiecutter(profile: Profile, answers: dict):
    """Call cookiecutter/cookieninja with the parameters from the TUI."""
    answers_profile = {}
    cookie_args = []
    questions = profile.config

    with open(COOKIECUTTER_FILE_PATH) as f:
        cookiecutter_config = json.load(f)

    # fill the answers with default value
    for question_id, question in questions.items():
        if not question.get("enabled", False) or question.get("control_flow"):
            continue

        if question.get("type") == "multiple-choices":
            for choice in question.get("choices", {}):
                choice_id = f"use_{choice.replace('-', '_')}"
                answers_profile[choice_id] = "no"
            continue

        answers_profile[question_id] = question.get(
            "default"
        ) or _get_cookiecutter_default_answer(cookiecutter_config[question_id])

    for question_id, answer in answers.items():
        if answer in [None, ""] or questions[question_id].get("control_flow"):
            continue

        if questions[question_id].get("type") != "multiple-choices":
            answers_profile[question_id] = answer
            continue

        for choice in answer:
            choice_id = f"use_{choice.replace('-', '_')}"
            answers_profile[choice_id] = "yes"

    for question_id, answer in answers_profile.items():
        cookie_args.append(f"{question_id}={answer}")

    sh_extras = {
        "_in": sys.stdin,
        "_out": sys.stdout,
        "_err": sys.stderr,
        "_no_err": True,
        "_env": os.environ,
        "_bg": True,
        "_bg_exc": False,
    }

    p = sh.cookieninja("--no-input", PACKAGE_PATH, *cookie_args, **sh_extras)

    try:
        p.wait()
    except sh.ErrorReturnCode as e:
        SciCookieLogs.raise_error(
            str(e), SciCookieErrorType.SH_ERROR_RETURN_CODE
        )
    except KeyboardInterrupt:
        pid = p.pid
        p.kill()
        SciCookieLogs.raise_error(
            f"Process {pid} killed.", SciCookieErrorType.SH_KEYBOARD_INTERRUPT
        )


def app():
    """Run SciCookie."""
    # note: this parameter should be provided by a CLI argument
    profile = Profile("osl")

    answers = make_questions(profile.config)

    if not answers:
        return

    call_cookiecutter(profile, answers)



#!/usr/bin/env python
import datetime
import os
import shutil
import subprocess

from pathlib import Path

PROJECT_DIRECTORY = Path(os.path.abspath(os.path.curdir)).resolve()

UNUSED_DOCS_DIRS = [
    PROJECT_DIRECTORY / 'docs-mkdocs',
    PROJECT_DIRECTORY / 'docs-sphinx',
    PROJECT_DIRECTORY / 'docs-jupyter-book'
]

DOCUMENTATION_ENGINE = "{{ cookiecutter.documentation_engine }}"
DOCS_SPEC_DIR = UNUSED_DOCS_DIRS.pop(
    UNUSED_DOCS_DIRS.index(
        PROJECT_DIRECTORY / f'docs-{DOCUMENTATION_ENGINE}'
    )
)

USE_SRC_LAYOUT = {{ cookiecutter.project_layout == "src" }}
if USE_SRC_LAYOUT:
    PACKAGE_PATH = PROJECT_DIRECTORY / "src" / "{{ cookiecutter.package_slug}}"
else:
    PACKAGE_PATH = PROJECT_DIRECTORY / "{{ cookiecutter.package_slug}}"

USE_BLACK = {{ cookiecutter.use_black == "yes" }}
USE_BLUE = {{ cookiecutter.use_blue == "yes" }}
USE_BANDIT = {{ cookiecutter.use_bandit == "yes" }}
USE_CONTAINERS = {{ cookiecutter.use_containers in ['Docker', 'Podman'] }}
USE_CLI = {{ cookiecutter.command_line_interface != "No command-line interface" }}
USE_CONDA = {{ cookiecutter.use_conda == "yes" }}
USE_MYPY = {{ cookiecutter.use_mypy == "yes" }}
{% if cookiecutter.code_of_conduct == "contributor-covenant" -%}
COC_PATH = PROJECT_DIRECTORY / 'coc' / 'CONTRIBUTOR_COVENANT.md'
{%- elif cookiecutter.code_of_conduct == "citizen-code-of-conduct" -%}
COC_PATH = PROJECT_DIRECTORY / 'coc' / 'CITIZEN.md'
{% else %}
COC_PATH = None
{%- endif %}
{% if cookiecutter.governance_document == "numpy-governance" -%}
GOVERNANCE_PATH = PROJECT_DIRECTORY / 'governance' / 'numpy_governance.md'
{% elif cookiecutter.code_of_conduct == "sciml-governance" -%}
GOVERNANCE_PATH = PROJECT_DIRECTORY / 'governance' / 'sciml_governance.md'
{% else -%}
GOVERNANCE_PATH = None
{%- endif %}
{% if cookiecutter.roadmap_document == "pytorch-ignite-roadmap" -%}
ROADMAP_PATH = PROJECT_DIRECTORY / 'roadmap' / 'ignite_roadmap.md'
{%- else %}
ROADMAP_PATH = None
{%- endif %}
{% if cookiecutter.build_system == "poetry" -%}
BUILD_SYSTEM = "poetry"
{% elif cookiecutter.build_system == "flit" -%}
BUILD_SYSTEM = "flit"
{% elif cookiecutter.build_system == "mesonpy" -%}
BUILD_SYSTEM = "mesonpy"
{% elif cookiecutter.build_system == "setuptools" -%}
BUILD_SYSTEM = "setuptools"
{% elif cookiecutter.build_system == "pdm" -%}
BUILD_SYSTEM = "pdm"
{% elif cookiecutter.build_system == "hatch" -%}
BUILD_SYSTEM = "hatch"
{% elif cookiecutter.build_system == "maturin" -%}
BUILD_SYSTEM = "maturin"
{% elif cookiecutter.build_system == "scikit-build-core" -%}
BUILD_SYSTEM = "scikit-build-core"
{% elif cookiecutter.build_system == "pybind11" -%}
BUILD_SYSTEM = "pybind11"
{%- else %}
BUILD_SYSTEM = None
{%- endif %}


def remove_dirs(dirs: list):
    for dirs in dirs:
        shutil.rmtree(dirs)


def remove_dir(dir_path):
    """Remove a directory located at PROJECT_DIRECTORY/dir_path"""
    shutil.rmtree(PROJECT_DIRECTORY/dir_path)


def remove_project_file(filepath: str):
    os.remove(PROJECT_DIRECTORY / filepath)


def remove_package_file(filepath: str):
    os.remove(PACKAGE_PATH / filepath)


def move_selected_doc_dir():
    docs_target_dir = PROJECT_DIRECTORY / "docs"
    for file_name in os.listdir(DOCS_SPEC_DIR):
        shutil.move(DOCS_SPEC_DIR / file_name, docs_target_dir)

    if DOCUMENTATION_ENGINE == "sphinx":
        remove_project_file(Path("docs") / "index.md")
        remove_project_file(Path("docs/api") / "references.md")

    shutil.rmtree(DOCS_SPEC_DIR)


def clean_up_docs():
    remove_dirs(UNUSED_DOCS_DIRS)
    move_selected_doc_dir()


def clean_up_project_layout():
    if USE_SRC_LAYOUT:
        if not os.path.exists("src"):
            os.mkdir("src")
            shutil.move('{{cookiecutter.package_slug}}', 'src')


def clean_up_code_of_conduct():
    if COC_PATH:
        shutil.move(
            COC_PATH,
            PROJECT_DIRECTORY / 'CODE_OF_CONDUCT.md'
        )
    remove_dir("coc")


def clean_up_conda():
    if not USE_CONDA:
        remove_dir("conda")


def clean_up_governance():
    if GOVERNANCE_PATH:
        shutil.move(
            GOVERNANCE_PATH,
            PROJECT_DIRECTORY / 'governance.md'
        )
    remove_dir("governance")


def clean_up_roadmap():
    if ROADMAP_PATH:
        shutil.move(
            ROADMAP_PATH,
            PROJECT_DIRECTORY / 'roadmap.md'
        )
    remove_dir("roadmap")


def clean_up_containers():
    if not USE_CONTAINERS:
        remove_dir("containers")


def clean_up_cli():
    if not USE_CLI:
        remove_package_file("__main__.py")


def clean_up_build_system():
    build_system_dir = PROJECT_DIRECTORY / "build-system"

    if BUILD_SYSTEM == "poetry":
        shutil.move(
            build_system_dir / "poetry-pyproject.toml",
            PROJECT_DIRECTORY / 'pyproject.toml'
        )
    elif BUILD_SYSTEM == "flit":
        shutil.move(
            build_system_dir / "flit-pyproject.toml",
            PROJECT_DIRECTORY / 'pyproject.toml'
        )
    elif BUILD_SYSTEM == "mesonpy":
        shutil.move(
            build_system_dir / "mesonpy-pyproject.toml",
            PROJECT_DIRECTORY / 'pyproject.toml'
        )
        shutil.move(
            build_system_dir / "meson.build",
            PROJECT_DIRECTORY / 'meson.build'
        )
    elif BUILD_SYSTEM == "setuptools":
        shutil.move(
            build_system_dir / "setuptools-pyproject.toml",
            PROJECT_DIRECTORY / 'pyproject.toml'
        )
    elif BUILD_SYSTEM == "pdm":
        shutil.move(
            build_system_dir / "pdm-pyproject.toml",
            PROJECT_DIRECTORY / 'pyproject.toml'
        )
    elif BUILD_SYSTEM == "hatch":
        shutil.move(
            build_system_dir / "hatch-pyproject.toml",
            PROJECT_DIRECTORY / 'pyproject.toml'
        )
    elif BUILD_SYSTEM == "maturin":
        shutil.move(
            build_system_dir / "maturin-pyproject.toml",
            PROJECT_DIRECTORY / 'pyproject.toml'
        )
        shutil.move(
            build_system_dir / "Cargo.toml",
            PROJECT_DIRECTORY / 'Cargo.toml'
        )
    elif BUILD_SYSTEM == "scikit-build-core":
        shutil.move(
            build_system_dir / "scikit-build-core-pyproject.toml",
            PROJECT_DIRECTORY / 'pyproject.toml'
        )
        shutil.move(
            build_system_dir / "CMakeLists.txt",
            PROJECT_DIRECTORY / 'CMakeLists.txt'
        )
        shutil.move(
            build_system_dir / "skcdemo.cpp",
            PROJECT_DIRECTORY / 'skcdemo.cpp'
        )
    elif BUILD_SYSTEM == "pybind11":
        shutil.move(
            build_system_dir / "pybind11-pyproject.toml",
            PROJECT_DIRECTORY / 'pyproject.toml'
        )
        shutil.move(
            build_system_dir / "CMakeLists.txt",
            PROJECT_DIRECTORY / 'CMakeLists.txt'
        )
        shutil.move(
            build_system_dir / "setup.py",
            PROJECT_DIRECTORY / 'setup.py'
        )                          
    else:
        shutil.move(
            build_system_dir / "base-pyproject.toml",
            PROJECT_DIRECTORY / 'pyproject.toml'
        )
    remove_dir("build-system")


def http2ssh(url):
    url = url.replace("https://", "git@")
    return url.replace("/", ":", 1)


def validation():
    if USE_BLUE and USE_BLACK:
        raise Exception(
            "The libs Blue and Black were selected, but you need to choose "
            "just one of them."
        )


def prepare_git():
    subprocess.call(["git", "init"])

    git_https_origin = http2ssh("{{cookiecutter.git_https_origin}}")
    git_https_upstream = http2ssh("{{cookiecutter.git_https_upstream}}")
    git_main_branch = http2ssh("{{cookiecutter.git_main_branch}}")
    unique_id = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    git_new_branch = f"add-initial-structure-{unique_id}"

    git_author_name = "{{cookiecutter.author_full_name}}"
    git_author_email = "{{cookiecutter.author_email}}"

    empty_repo = not (git_https_origin or git_https_upstream)

    if not empty_repo:
        subprocess.call(["git", "stash"])

    if git_https_origin != "":
        subprocess.call(["git", "remote", "add", "origin", git_https_origin])
        subprocess.call(["git", "fetch", "--all"])
        subprocess.call(["git", "checkout", f"origin/{git_main_branch}"])

    if git_https_upstream != "":
        subprocess.call(
            ["git", "remote", "add", "upstream", git_https_upstream]
        )
        subprocess.call(["git", "fetch", "--all"])

    subprocess.call(
        ["git", "config", "user.name", git_author_name]
    )
    subprocess.call(
        ["git", "config", "user.email", git_author_email]
    )

    subprocess.call(["git", "checkout", "-b", git_new_branch])

    if not empty_repo:
        subprocess.call(["git", "stash", "pop"])

    subprocess.call(["git", "add", "."])
    subprocess.call(["git", "commit", "-m", "Initial commit", "--no-verify"])


def add_binding_source_files():
    if BUILD_SYSTEM == "maturin":
        build_system_dir = PROJECT_DIRECTORY / "build-system"
        src_system_dir = PROJECT_DIRECTORY/ "src"
        if USE_SRC_LAYOUT :
            shutil.move(build_system_dir / "lib.rs", "src")
        else:
            os.makedir(src_system_dir)
            shutil.move(build_system_dir / "lib.rs", src_system_dir)
    elif BUILD_SYSTEM == "pybind11" :
        build_system_dir = PROJECT_DIRECTORY / "build-system"
        src_system_dir = PROJECT_DIRECTORY/ "src"
        if USE_SRC_LAYOUT :
            shutil.move(build_system_dir / "main.cpp", "src")
        else:
            os.makedir(src_system_dir)
            shutil.move(build_system_dir / "main.cpp", src_system_dir)
    else:
        pass


def clean_up_mypy():
    if not USE_MYPY:
        remove_package_file("py.typed")


def post_gen():
    validation()

    # keep this one first, because it changes the package folder
    clean_up_project_layout()
    add_binding_source_files()
    clean_up_cli()
    clean_up_mypy()
    clean_up_code_of_conduct()
    clean_up_conda()
    clean_up_containers()
    clean_up_docs()
    clean_up_governance()
    clean_up_roadmap()
    clean_up_build_system()

    # keep it at the end, because it will create a new git commit
    prepare_git()


if __name__ == "__main__":
    post_gen()



"""Tests for {{ cookiecutter.package_slug }} package."""
{% if cookiecutter.use_pytest == "yes" -%}
import pytest
{% endif -%}
{% if cookiecutter.use_hypothesis == "yes" -%}
from hypothesis import given
from hypothesis import strategies as st
{% endif -%}
{% if cookiecutter.use_pytest == "yes" %}

@pytest.fixture
def response_pytest():
    """Sample pytest fixture."""
    return True
{%- if cookiecutter.use_hypothesis == "yes" %}


@pytest.fixture
def response_hypothesis():
    """Sample pytest + hypothesis fixture."""
    return True
{%- endif -%}
{%- endif -%}

{%- if cookiecutter.use_pytest == "yes" %}


def test_content_pytest():
    """Test with pytest."""
    assert True
{%- if cookiecutter.use_hypothesis == "yes" %}


@given(st.text())
def test_content_hypothesis(response_hypothesis):
    """Test with pytest + hypothesis."""
    assert response_hypothesis
{%- endif -%}
{%- endif %}
{# keep this line at the end of the file -#}
